/*
	Client and event management routines.

	Created by Adrian Robinson.
*/
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>

#include <sys/proxy.h>
#include <sys/kernel.h>

#include "local.h"

static MouseServer *sMSList;


static MouseServer *openMouseServer(nid_t nid)
{
	MouseServer *server = sMSList;

	while (server) {
		if (server->nid == nid)
			break;

		server = server->next;
	}

	if (server == NULL) {
		struct mouse_event empty;

		if ((server = (MouseServer *)calloc(sizeof(*server), 1)) == NULL)
			return NULL;

		/* -------- attempt to connect to mouse server on client's node ------- */
		if ((server->mc = mouse_open(nid, 0, -1)) == NULL)
			return NULL;
	
		/* -------- flush any pending events --------- */
		mouse_flush(server->mc);
	
		server->vmproxy = server->mproxy = qnx_proxy_attach(0, 0, 0, -1);
		if (nid != getnid())
			server->vmproxy = qnx_proxy_rem_attach(nid, server->mproxy);
	
		/* -------- arm mouse proxy by reading zero bytes --------- */
		mouse_read(server->mc, &empty, 0, server->vmproxy, 0);

		server->next = sMSList;
		server->nid  = nid;
		sMSList = server;
	}

	server->connectCount++;
	return server;
}



static void closeMouseServer(MouseServer *server)
{
	if (server->connectCount)
		server->connectCount--;
}
	


int findClient(pid_t pid, int unused, int proxy)
{
	register MouseClient far *client = eMMStats->clients;
	register int count;
	int max = eMMStats->maxClients;

	if (!unused && eMMStats->clientCount == 0)
		return -1;

	for (count = 0; count < max; count++, client++) {
		if (unused) {
			if (client->pid == 0)
				break;
		}
		else if (proxy) {
			if (client->pid && client->server->mproxy == pid)
				break;
		}
		else if (client->pid == pid)
			break;
	}

	return (count < eMMStats->maxClients ? count : -1);
}



int addClient(pid_t pid, struct _mouseman_open *msg, int *handle)
{
	register MouseClient far *client;
	struct _event_list mm;

	if ((*handle = findClient(0, 1, 0)) == -1)
		return EMFILE;

	client = &eMMStats->clients[*handle];

	if ((client->server = openMouseServer(msg->node)) == NULL)
		return errno;

	/* -------- inform other registered clients of new client -------- */
	mm.node.type = _NEWCLIENT;
	mm.node.nid  = msg->node;
	distributeEvent(&mm);

	client->pid    = pid;
	client->eproxy = msg->eproxy;
	client->nid    = msg->node;

	eMMStats->clientCount++;
	return EOK;
}


int delClient(pid_t pid, int handle)
{
	register MouseClient far *client;
	register struct _event_list *event, *prev;
	struct _event_list mm;
	nid_t node;
	
	if (handle == -1) {
		if ((handle = findClient(pid, 0, 0)) == -1)
			return ESRCH;
	}
	else if (handle > eMMStats->maxClients)
		return EBADF;

	client = &eMMStats->clients[handle];

	if (client->pid != pid)
		return EPERM;
	node = client->nid;

	/* ------ step through and free event list not read by client ----- */
	event = client->start;
	while (event) {
		prev = event;
		event = event->next;
		free(prev);
	}

	closeMouseServer(client->server);
	client->pid = 0;
	eMMStats->clientCount--;

	/* -------- inform other registered clients of client death -------- */
	mm.node.type = _DELCLIENT;
	mm.node.nid  = node;
	distributeEvent(&mm);

	return EOK;
}



/* ================ event management routines ============== */

static void distributeEvent(struct _event_list *ev)
{
	register MouseClient far *client = eMMStats->clients;
	register int count;
	struct _event_list *mm;
	int max = eMMStats->maxClients;
	pid_t expect;
	
	if (eMMStats->clientCount == 0)
		return;

	/* ----------- add event to all clients currently registered -------- */
	for (count = 0; count < max; count++, client++) {
		if (client->pid == 0)
			continue;

		if ((mm = (struct _event_list *)calloc(sizeof(*mm), 1)) == NULL)
			return;

		mm->node = ev->node;

		/* ------- add event to this client's list ------ */
		if (client->end == NULL)
			client->start = mm;
		else
            client->end->next = mm;
		client->end = mm;
		client->eventCount++;

		/* --- notify client of event and check that pid is still around --- */
		if ((expect = Trigger(client->eproxy)) != client->pid && expect) 
			delClient(client->pid, count);
	}
}


void addEvent(int handle)
{
	register MouseClient far *client = &eMMStats->clients[handle];
	struct _event_list mm;
	int empty;
	
	do {
		if (mouse_read(client->server->mc, &mm.node.event, 1, client->server->vmproxy, &empty)) {
			mm.node.type = _EVENT;
			mm.node.nid  = client->nid;
			distributeEvent(&mm);
		}
	} while (!empty);
}


int readEvent(pid_t pid, int handle,
			  struct _mouseman_read_reply *rply,  
			  size_t *size)
{
	register MouseClient far *client;
	register int count;
	struct _event_list *prev;
	int index, events;

	if ((index = findClient(pid, 0, 0)) == -1)
		return ESRCH;

	if (index != handle)
		return EPERM;
	
	client = &eMMStats->clients[handle];

	/* ------- determine the # of events to return ------ */
	events = min(client->eventCount, MAX_EVENTS);

	for (count = 0; count < events; count++) {
		memcpy(&rply->events[count], &client->start->node, sizeof(prev->node));
		prev = client->start;
		client->start = client->start->next;
		free(prev);
	}

	client->eventCount -= events;
	rply->count = events;
	if (rply->empty = client->eventCount == 0)
		client->start = client->end = NULL;

	*size = sizeof(*rply) - ((MAX_EVENTS - events) * sizeof(rply->events[0]));
	return EOK;
}
